/*[clinic input]
preserve
[clinic start generated code]*/

PyDoc_STRVAR(_testlimitedcapi_call_vectorcall__doc__,
"call_vectorcall($module, callable, /)\n"
"--\n"
"\n");

#define _TESTLIMITEDCAPI_CALL_VECTORCALL_METHODDEF    \
    {"call_vectorcall", (PyCFunction)_testlimitedcapi_call_vectorcall, METH_O, _testlimitedcapi_call_vectorcall__doc__},

PyDoc_STRVAR(_testlimitedcapi_call_vectorcall_method__doc__,
"call_vectorcall_method($module, callable, /)\n"
"--\n"
"\n");

#define _TESTLIMITEDCAPI_CALL_VECTORCALL_METHOD_METHODDEF    \
    {"call_vectorcall_method", (PyCFunction)_testlimitedcapi_call_vectorcall_method, METH_O, _testlimitedcapi_call_vectorcall_method__doc__},
/*[clinic end generated code: output=5976b9b360e1ff30 input=a9049054013a1b77]*/
